#include <gtk/gtk.h>
#include <stdio.h>

#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>
#include <xmms/configfile.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#ifndef VERSION
# define VERSION "0.3.0"
#endif

#ifdef PLUGIN_DEBUG
#define OUT( exp ) exp
#else
#define OUT( exp )
#endif

#define SCORE_DIR -314159

#ifdef PLUGIN_DEBUG
# define SLEEP usleep(100)
#else
# ifdef HAVE_PTHREAD_H 
#   define SLEEP pthread_yield()
# else
#   ifdef _POSIX_PRIORITY_SCHEDULING
#     include <sched.h>
#     define SLEEP sched_yield()
#   else
#     define SLEEP
#   endif
# endif
#endif

static void init(void);
static void cleanup(void);
static gint timeout_func(gpointer);
static void clean_list( void );
static void configure( void );
static gint timeout_tag = 0;
static gint previous_song = -1, previous_length = -2, old_cpos = 0, old_length = 0, last_song = -1, last_titles = -1;
static gint was_stopped = TRUE;
static gboolean avoid_twice = FALSE;
static gdouble medium_score = 0;
static gchar *score_file = NULL;

GtkWidget *configure_win = NULL;
GtkWidget *check;

GeneralPlugin sc_gp =
{
	NULL,			/* handle */
	NULL,			/* filename */
	-1,			/* xmms_session */
	NULL,			/* Description */
	init,
	NULL,
	configure,
	cleanup,
};

typedef struct
{
    gint  number;
    gchar played;
    gint  score;
} TitleStatus;

GSList *title_list;

GeneralPlugin *get_gplugin_info(void)
{
	sc_gp.description = g_strdup_printf( "Real Random %s", VERSION );
	return &sc_gp;
}

static gint get_inode( gchar *file )
{
    struct stat buf;

    if( stat( file, &buf ) == 0 )
    {
	if( S_ISDIR( buf.st_mode ) )
	    return -1;

	return buf.st_ino;
    }

    return 0;
}


static void read_config( void )
{
    ConfigFile *cfgfile;

    if ((cfgfile = xmms_cfg_open_default_file()) != NULL)
    {
	if( !xmms_cfg_read_boolean(cfgfile, "real_random", "avoid_twice", &avoid_twice ) )
	    avoid_twice = TRUE;
	
	xmms_cfg_free(cfgfile);
    }
}

static void init(void)
{
	previous_song = -1;
	previous_length = -2;	
	old_cpos = 0;
	old_length = 0;
	last_song = -1;
	
	srand( time(NULL) );
	title_list = NULL;
	was_stopped = TRUE;

	score_file = g_strconcat( g_get_home_dir(), "/.xmms/score", NULL );

	read_config();

	timeout_tag = gtk_timeout_add(100, timeout_func, NULL);
}

static void cleanup(void)
{
	if (timeout_tag)
		gtk_timeout_remove(timeout_tag);
	timeout_tag = 0;

	clean_list();
	g_free( score_file );
}

static void clean_list( void )
{
	if( title_list )
	{
	    GSList *node;

	    for( node = title_list; node; node =  node->next )
		g_free( node->data );

	    g_slist_free( title_list );
	    title_list = NULL;
	}
}

static gint get_score_of_file( ConfigFile *cfgfile, gchar *file )
{

    gint score;
    gint inode;
    gchar *str;

    if( !cfgfile )
	return 0;

    inode = get_inode( file );

    if( !inode )
	return 0; 

    if( inode == -1 )
	return SCORE_DIR; // skip directories or things like this   

    str = g_strdup_printf( "%d", inode );
    
    if( xmms_cfg_read_int( cfgfile, "score", str , &score ) )   
	return score;
   
    g_free( str );

    return 0;
}

static void create_list( void )
{
    gint length = xmms_remote_get_playlist_length( sc_gp.xmms_session );
    gint i;
    ConfigFile *cfgfile; 

    cfgfile = xmms_cfg_open_file( score_file );

    medium_score = 0;

    for( i = 0; i < length; ++i )
    {
	gchar *file;

	TitleStatus *track = g_malloc( sizeof( TitleStatus ) );

	track->number = i;
	track->played = FALSE;
	file = xmms_remote_get_playlist_file( sc_gp.xmms_session, i );
	track->score = get_score_of_file( cfgfile, file );
	g_free( file );

	title_list = g_slist_append( title_list, track );

	if( track->score != SCORE_DIR )
	    medium_score += track->score;
    }

    if( cfgfile )
	xmms_cfg_free( cfgfile );

    if( length )
	medium_score /= (gdouble)length;

    printf( "medium score: %f\n", medium_score );
}

static void inc_score( gint track, gint inc )
{
    GSList *node;
    gint i;
    gdouble length;
    
    length =  xmms_remote_get_playlist_length( sc_gp.xmms_session );

    for( i = 0, node = title_list; node && i < track + 1; i++, node = node->next )
	if( i == track )
	{
	    gchar *file;
	    ConfigFile *cfgfile;

	    ((TitleStatus*)node->data)->score += inc;

	    medium_score += (gdouble)(((TitleStatus*)node->data)->score) / length;

	    printf( "medium score: %f\n", medium_score );

	    file = xmms_remote_get_playlist_file( sc_gp.xmms_session, i );
	    cfgfile = xmms_cfg_open_file( score_file  );

	    if( !cfgfile )
		cfgfile = xmms_cfg_new();

	    if( cfgfile )
	    {	
		gchar *str;

		gint inode = get_inode( file );

		str = g_strdup_printf( "%d", inode );

		xmms_cfg_write_int( cfgfile, "score", str , ((TitleStatus*)node->data)->score );
		xmms_cfg_write_file( cfgfile,  score_file );
		xmms_cfg_free( cfgfile );
		g_free( str );
	    }
	}
}

static gint get_next_random_track( void )
{
    gint length;
    gint npos;
    GSList *node;
    gint i;


    if( !title_list )
	return 0;
    
    length = g_slist_length( title_list );

    if(  length < 2 )
	return 0;

    npos = (int) ((double)length * rand() / RAND_MAX + 1.0 );
    
    if( !avoid_twice )
	return npos;

    node = g_slist_nth( title_list, npos );
    
    // find first free entry from current pos
    while( node && ((TitleStatus*)node->data)->played && (((TitleStatus*)node->data)->score >= medium_score ) )
    {
	node = node->next;
    }
    
    if( !node )
    {
	// find any free entry
	
	i = 0;
	for( node = title_list; node; node = node->next )
	{
	    if( !((TitleStatus*)node->data)->played && (((TitleStatus*)node->data)->score >= medium_score ))
		break;
	    
	    if( i > npos )
		break;
	    
	    i++;
	}
    }
    
    
    if( !node )
    {
	// if we still have no playable title all titles were played
    
	g_message( "all tracks played, list recreated\n" );

	clean_list();
	create_list();
	
	return get_next_random_track();
    }

    ((TitleStatus*)node->data)->played = TRUE;

    return ((TitleStatus*)node->data)->number;
}

static gint timeout_func(gpointer data)
{
        gint pos, length, cpos, titles;
	gboolean playing, by_plugin;

	playing = xmms_remote_is_playing(sc_gp.xmms_session);

	if( !xmms_remote_is_shuffle( sc_gp.xmms_session ) || !playing )
	{
	    if( !was_stopped )
		was_stopped = TRUE;
	    was_stopped++;

	    if( was_stopped > 5 )
	    {
		previous_song = -1;
		previous_length = -2;
		last_song = -1;
		old_cpos = 0;
		old_length = 0;
		last_titles = -1;
	    }

	    return TRUE;
	}

	pos = xmms_remote_get_playlist_pos(sc_gp.xmms_session);
	length = xmms_remote_get_playlist_time(sc_gp.xmms_session, pos );
	cpos = xmms_remote_get_output_time( sc_gp.xmms_session);	
	titles = xmms_remote_get_playlist_length( sc_gp.xmms_session );

	if( previous_song == -1 )
	{
	    previous_song = pos;
	    previous_length = length;
	    last_song = -1;
	    old_cpos = 0;
	    old_length = 0;
	    last_titles = titles;

	    printf( "resetting\n" );
	}

	if (playing)
	{
	    gboolean song_changed = (old_length != length);
	    
	    OUT( printf( "ps: %d pos: %d length: %d cpos: %d title: %d\n", last_song, pos, length, cpos, titles ) );

	    if( was_stopped )
	    {
		OUT( printf( "was stopped\n" ) );
	    }

	    if( !title_list || (last_titles != titles) )
	    {
		OUT( printf( "createing list\n" ) );

		clean_list();
		create_list();

		last_titles = titles;
	    }
	    
	    if( pos != previous_song && old_length - old_cpos < 1000 )
	    {
		gint npos, wait = 0;

		g_message( "updating playlist\n" );

		g_message( "current pos: %d\n", pos );

		npos = get_next_random_track();

		printf( "new pos: %d\n", npos );

		xmms_remote_set_playlist_pos( sc_gp.xmms_session, npos );		

		while(  xmms_remote_get_playlist_pos(sc_gp.xmms_session) != npos )
		{
		    SLEEP;
		    wait++;

		    if( wait > 3000 )
			break;
		}


		last_song = pos;
		previous_song = npos;
		previous_length = length;

		by_plugin = TRUE;
		song_changed = TRUE; 

		inc_score( pos, 1 );

	    } else {
		
		last_song = pos;
		old_cpos = cpos;
		old_length = length;

		by_plugin = FALSE;

	    }

	    if( song_changed )
	    {
		printf( "song changed; last song: %d new song: %d\n", last_song, pos );

		if( ((was_stopped < 5) && !by_plugin ) && !(previous_song == pos))
		{
		    inc_score( last_song, -1 );
		    printf( "manually %d\n", was_stopped );
		}		
	    }

	    was_stopped = FALSE;

	} 

	return TRUE;
}

static void ok_cb(gpointer data)
{
	ConfigFile *cfgfile;	

	avoid_twice = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( check ) );

	cfgfile = xmms_cfg_open_default_file();
	xmms_cfg_write_boolean(cfgfile, "real_random", "avoid_twice", avoid_twice );
	xmms_cfg_write_default_file(cfgfile);
	xmms_cfg_free(cfgfile);

	gtk_widget_destroy(configure_win);
	configure_win = NULL;
}


static void cancel_cb(gpointer data)
{
	gtk_widget_destroy(configure_win);
	configure_win = NULL;
}


static void configure( void )
{
    GtkWidget *vbox;
    GtkWidget *label;
    GtkWidget *ok;
    GtkWidget *cancel;
    GtkWidget *hbox;

    if( configure_win )
	return;

    configure_win = gtk_window_new( GTK_WINDOW_DIALOG );
    gtk_window_set_title( GTK_WINDOW( configure_win ), "Real Random Configuration" );

    gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);

    vbox = gtk_vbox_new( FALSE, 10 );
    gtk_container_add( GTK_CONTAINER( configure_win ), vbox );

    label = gtk_label_new( "Real Random Plugin\nThe plugin is only active if the shuffle mode is turned \"on\"." );
    gtk_container_add( GTK_CONTAINER( vbox ), label );

    check = gtk_check_button_new_with_label( "avoid playing a title twice" );
    gtk_container_add( GTK_CONTAINER( vbox ), check );
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( check ), avoid_twice );

    label = gtk_label_new( "(C) by Martin Oberzalek 2003" );
    gtk_container_add( GTK_CONTAINER( vbox ), label );

    hbox = gtk_hbutton_box_new();
    gtk_container_add( GTK_CONTAINER( vbox ), hbox );
    
    ok = gtk_button_new_with_label( "Ok" );
    gtk_box_pack_start(GTK_BOX(hbox), ok, TRUE, TRUE, 0);
    gtk_signal_connect( GTK_OBJECT( ok ), "clicked", GTK_SIGNAL_FUNC( ok_cb ), NULL );
    GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(ok);


    cancel = gtk_button_new_with_label( "Cancel" );
    gtk_signal_connect( GTK_OBJECT( cancel ), "clicked", GTK_SIGNAL_FUNC( cancel_cb ), NULL );
    gtk_box_pack_start(GTK_BOX(hbox), cancel, TRUE, TRUE, 0);

    gtk_widget_show_all( configure_win );

}

#ifdef PLUGIN_DEBUG

int main( int argc, char ** argv )
{
    gtk_init( &argc, &argv );

    GeneralPlugin *gp = get_gplugin_info();

    gp->xmms_session = 0;

    gp->init();

    gtk_main();

    gp->cleanup();

    return 0;
}

#endif
